'use strict';

const weekdayShort = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"];
const weekdayLong = ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];
const monthShort = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
const monthLong = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
const en_US = {
    "%a": dt => weekdayShort[dt.getDay()],
    "%A": dt => weekdayLong[dt.getDay()],
    "%b": dt => monthShort[dt.getMonth()],
    "%B": dt => monthLong[dt.getMonth()],
    "%p": dt => (dt.getHours() < 12 ? "AM" : "PM"),
    // Sun, Jan 2, 2022, 3:04:05 AM
    "%c": "%a, %b %-d, %Y, %-I:%M:%S %p",
    // 3:04:05 AM
    "%r": "%-I:%M:%S %p",
    // 1/2/22
    "%x": "%-m/%-d/%y",
    // 3:04:05 AM
    "%X": "%-I:%M:%S %p",
};

const strftimeHandlers = () => {
    const pad2 = fn => (dt) => ("0" + fn(dt)).substr(-2);
    const pad2S = fn => (dt) => (" " + fn(dt)).substr(-2);
    const pad3 = fn => (dt) => ("00" + fn(dt)).substr(-3);
    const padY = fn => (dt) => {
        let year = fn(dt);
        if (0 <= year && year <= 9999) {
            return ("000" + year).substr(-4);
        }
        let prefix = "+";
        if (year < 0) {
            prefix = "-";
            year = -year;
        }
        return prefix + (("00000" + year).substr(-6));
    };
    /**
     * %C     The century number (year/100) as a 2-digit integer. (SU)
     */
    const C = dt => Math.floor(dt.getFullYear() / 100);
    /**
     * %d     The day of the month as a decimal number (range 01 to 31).
     */
    const d = dt => dt.getDate();
    /**
     * %H     The hour as a decimal number using a 24-hour clock (range 00 to 23).
     */
    const H = dt => dt.getHours();
    /**
     * %I     The hour as a decimal number using a 12-hour clock (range 01 to 12).
     */
    const I = dt => (((dt.getHours() + 11) % 12) + 1);
    /***
     * %L - Millisecond of the second (000..999)
     */
    const L = dt => dt.getMilliseconds();
    /**
     * %m     The month as a decimal number (range 01 to 12).
     */
    const m = dt => (dt.getMonth() + 1);
    /**
     * %M     The minute as a decimal number (range 00 to 59).
     */
    const M = dt => dt.getMinutes();
    /**
     * %S     The second as a decimal number (range 00 to 60)
     */
    const S = dt => dt.getSeconds();
    /**
     * %U     The week number [NOT IMPLEMENTED]
     * %y     The year as a decimal number without a century (range 00 to 99).
     */
    const y = dt => dt.getFullYear() % 100;
    /**
     * %Y     The year as a decimal number including the century.
     */
    const Y = dt => dt.getFullYear();
    /**
     * %z     The +hhmm or -hhmm numeric timezone (that is, the hour and minute offset from UTC). (SU)
     */
    const makeZ = (delim) => {
        return dt => {
            let offset = -dt.getTimezoneOffset();
            const isMinus = (offset < 0);
            if (isMinus)
                offset = -offset;
            const hour = Math.floor(offset / 60);
            const min = offset % 60;
            return (isMinus ? "-" : "+") + (hour < 10 ? "0" + hour : hour) + delim + (min < 10 ? "0" + min : min);
        };
    };
    /**
     * %c %r %x and %X are defined at locale files
     */
    return {
        "%c": "%a %b %e %T %Y",
        "%-C": C,
        "%C": pad2(C),
        "%-d": d,
        "%d": pad2(d),
        "%D": "%m/%d/%y",
        "%e": pad2S(d),
        "%F": "%Y-%m-%d",
        "%-H": H,
        "%H": pad2(H),
        "%-I": I,
        "%I": pad2(I),
        "%k": pad2S(H),
        "%l": pad2S(I),
        "%-L": pad3(L),
        "%L": pad3(L),
        "%-m": m,
        "%m": pad2(m),
        "%-M": M,
        "%M": pad2(M),
        "%P": dt => (dt.getHours() < 12 ? "am" : "pm"),
        "%R": "%H:%M",
        "%-S": S,
        "%S": pad2(S),
        "%T": "%H:%M:%S",
        "%-y": y,
        "%y": pad2(y),
        "%-Y": Y,
        "%Y": padY(Y),
        "%v": "%e-%b-%Y",
        "%w": dt => dt.getDay(),
        "%:z": makeZ(":"),
        "%z": makeZ(""),
        "%%": () => "%",
        "%n": () => "\n",
        "%t": () => "\t", // %t     A tab character. (SU)
    };
};

const formatHandlers = {
    YY: "%y",
    YYYY: "%Y",
    M: "%-m",
    MM: "%m",
    MMM: "%b",
    MMMM: "%B",
    D: "%-d",
    DD: "%d",
    d: "%w",
    dd: "%a",
    ddd: "%a",
    dddd: "%A",
    H: "%-H",
    HH: "%H",
    h: "%-I",
    hh: "%I",
    m: "%-M",
    mm: "%M",
    s: "%-S",
    ss: "%S",
    SSS: "%L",
    Z: "%:z",
    ZZ: "%z",
    A: "%p",
    a: "%P", // am pm
};

const mergeRouter = (a, b) => ((a && b) ? (specifier => (a(specifier) || b(specifier))) : (a || b));
const makeRouter = (handlers) => (handlers && (specifier => handlers[specifier]));
// @see https://docs.ruby-lang.org/en/3.1/DateTime.html#method-i-strftime
const strftimeRE = /%(?:[EO]\w|[0_#^-]?[1-9]?\w|::?z|[%+])/g;
// /\[(.*?)\]|A+|a+|B+|b+|C+|c+|...|Z+|z+/g
const makeFormatRE = () => {
    let re = ["\\[(.*?)\\]"];
    const c = (code) => String.fromCharCode(code + 65) + "+";
    for (let i = 0; i < 26; i++) {
        re.push(c(i), c(i + 32));
    }
    return new RegExp(re.join("|"), "g");
};
const formatRE = makeFormatRE();
const ISO = { ISO: "%Y-%m-%dT%H:%M:%S.%L%:z" };
const makeTexter = (router) => {
    const one = (specifier, dt) => {
        let handler = router(specifier);
        if ("string" === typeof handler) {
            const next = router(handler);
            if (next != null)
                handler = next; // bypass strftime call
        }
        if ("function" === typeof handler) {
            return handler(dt);
        }
        else if (handler == null) {
            return specifier; // Unsupported specifiers
        }
        else {
            return strftime(handler, dt); // recursive call
        }
    };
    const out = {};
    const strftime = out.strftime = (fmt, dt) => {
        if (fmt == null)
            return one("ISO", dt);
        return fmt.replace(strftimeRE, (specifier) => one(specifier, dt));
    };
    out.format = (fmt, dt) => {
        if (fmt == null)
            return one("ISO", dt);
        return fmt.replace(formatRE, (specifier, raw) => (raw || one(specifier, dt)));
    };
    out.handler = specifiers => makeTexter(mergeRouter(makeRouter(specifiers), router));
    return out;
};
const texter = makeTexter().handler(ISO).handler(en_US).handler(formatHandlers).handler(strftimeHandlers());
const _strftime = texter.strftime;
const strftime = (fmt, dt) => _strftime(fmt, dt || new Date());
const getTexter = (x) => (x && x.tx || texter);
const formatPlugin = (Parent) => {
    return class CDateFormat extends Parent {
        /**
         * updates strftime option with the given locale
         */
        handler(handlers) {
            const out = this.inherit();
            const { x } = out;
            x.tx = getTexter(x).handler(handlers);
            return out;
        }
        /**
         * returns a text with "YYYY-MM-DD" formatting style
         */
        format(fmt) {
            return getTexter(this.x).format(fmt, this.ro());
        }
        /**
         * returns a text with "%y/%m/%d formatting style
         */
        text(fmt) {
            return getTexter(this.x).strftime(fmt, this.ro());
        }
    };
};

const unitMap = {
    year: "y" /* Unit.year */,
    month: "M" /* Unit.month */,
    week: "w" /* Unit.week */,
    date: "d" /* Unit.day */,
    day: "d" /* Unit.day */,
    hour: "h" /* Unit.hour */,
    minute: "m" /* Unit.minute */,
    second: "s" /* Unit.second */,
    millisecond: "ms" /* Unit.millisecond */,
};
Object.keys(unitMap).forEach((key) => {
    const s = (key + "s");
    const v = unitMap[key];
    unitMap[s] = unitMap[v] = unitMap[key];
});
const getUnitShort = (unit) => {
    const u = unitMap[(unit || "ms" /* Unit.millisecond */)] ||
        unitMap[String(unit).toLowerCase()];
    if (!u)
        throw new RangeError("Invalid unit: " + unit);
    return u;
};
const unitMS = {
    h: 3600000 /* d.HOUR */,
    m: 60000 /* d.MINUTE */,
    s: 1000 /* d.SECOND */,
    ms: 1,
};

const addMonth = (dt, months) => {
    const year = dt.getFullYear();
    const month = dt.getMonth();
    const date = dt.getDate();
    // calculate days between the months
    const tmp = new Date(year, month, 1);
    const before = +tmp;
    const diff = year * 12 + month + months;
    tmp.setFullYear(Math.floor(diff / 12));
    const newMonth = diff % 12;
    tmp.setMonth(newMonth);
    const days = Math.round((+tmp - before) / 86400000 /* d.DAY */);
    // move days
    addDay(dt, days);
    // check an overflow
    const newDate = dt.getDate();
    if (newMonth !== dt.getMonth() && date > newDate) {
        // the very last day of the previous month
        addDay(dt, -newDate);
    }
};
const addDay = (dt, days) => {
    const tz1 = dt.getTimezoneOffset();
    dt.setTime(+dt + days * 86400000 /* d.DAY */);
    const tz2 = dt.getTimezoneOffset();
    // adjustment for Daylight Saving Time (DST)
    if (tz1 !== tz2) {
        dt.setTime(+dt + (tz2 - tz1) * 60000 /* d.MINUTE */);
    }
};
const add = (dt, diff, unit) => {
    if (!diff)
        return;
    const u = getUnitShort(unit);
    const msec = unitMS[u];
    if (msec) {
        dt.setTime(+dt + diff * msec);
        return;
    }
    switch (u) {
        case "y" /* Unit.year */:
            return addMonth(dt, diff * 12);
        case "M" /* Unit.month */:
            return addMonth(dt, diff);
        case "w" /* Unit.week */:
            return addDay(dt, diff * 7);
        case "d" /* Unit.day */:
            return addDay(dt, diff);
    }
};

const startOfMonth = (dt) => {
    startOfDay(dt);
    add(dt, 1 - dt.getDate(), "day");
};
const startOfDay = (dt) => {
    const tz1 = dt.getTimezoneOffset();
    truncate(dt, 86400000 /* d.DAY */);
    const tz2 = dt.getTimezoneOffset();
    // adjustment for Daylight Saving Time (DST)
    if (tz1 !== tz2) {
        dt.setTime(+dt + (tz2 - tz1) * 60000 /* d.MINUTE */);
    }
};
const truncate = (dt, unit) => {
    const tz = dt.getTimezoneOffset() * 60000 /* d.MINUTE */;
    dt.setTime(Math.trunc((+dt - tz) / unit) * unit + tz);
};
const startOf = (dt, unit) => {
    const u = getUnitShort(unit);
    const msec = unitMS[u];
    if (msec)
        return truncate(dt, msec);
    switch (u) {
        case "y" /* Unit.year */:
            startOfMonth(dt);
            return add(dt, -dt.getMonth(), "M" /* Unit.month */);
        case "M" /* Unit.month */:
            return startOfMonth(dt);
        case "w" /* Unit.week */:
            startOfDay(dt);
            return add(dt, -dt.getDay(), "d" /* Unit.day */);
        case "d" /* Unit.day */:
            return startOfDay(dt);
    }
};

const calcPlugin = (Parent) => {
    return class CDateCalc extends Parent {
        /**
         * returns a new CDate object manipulated
         */
        startOf(unit) {
            const dt = this.rw();
            startOf(dt, unit);
            return this.create(dt);
        }
        /**
         * returns a new CDate object manipulated
         */
        endOf(unit) {
            const dt = this.rw();
            startOf(dt, unit);
            add(dt, 1, unit);
            add(dt, -1);
            return this.create(dt);
        }
        /**
         * returns a new CDate object manipulated
         */
        add(diff, unit) {
            const dt = this.rw();
            add(dt, diff, unit);
            return this.create(dt);
        }
        /**
         * returns a new CDate object manipulated
         */
        next(unit) {
            return this.add(1, unit);
        }
        /**
         * returns a new CDate object manipulated
         */
        prev(unit) {
            return this.add(-1, unit);
        }
    };
};

class DateUTC {
    constructor(dt) {
        this.dt = dt;
        //
    }
    valueOf() {
        return +this.dt;
    }
    setTime(msec) {
        return this.dt.setTime(msec);
    }
    getTimezoneOffset() {
        return 0; // always UTC
    }
    getMilliseconds() {
        return this.dt.getUTCMilliseconds();
    }
    getSeconds() {
        return this.dt.getUTCSeconds();
    }
    getMinutes() {
        return this.dt.getUTCMinutes();
    }
    getHours() {
        return this.dt.getUTCHours();
    }
    getDay() {
        return this.dt.getUTCDay();
    }
    getDate() {
        return this.dt.getUTCDate();
    }
    ;
    getMonth() {
        return this.dt.getUTCMonth();
    }
    getFullYear() {
        return this.dt.getUTCFullYear();
    }
    getTime() {
        return +this;
    }
}
const utcPlugin = (Parent) => {
    return class CDateUTC extends Parent {
        utc() {
            const out = this.inherit();
            out.x.rw = (dt) => new DateUTC(new Date(+dt));
            return out;
        }
    };
};

/**
 * lazy build
 */
const lazy = (fn) => {
    let data;
    return () => (data || (data = fn()));
};
/**
 * cache
 */
const cached = (fn) => {
    let cached = {};
    return key => (cached[key] || (cached[key] = fn(key)));
};

const parseTZ = cached((tz) => {
    const matched = tz.match(/(?:^|GMT)?(?:([+-])([01]?\d):?(\d[05])|$)|UTC$/);
    if (!matched)
        return;
    const offset = ((+matched[2]) * 60 + (+matched[3])) | 0;
    return (matched[1] === "-") ? -offset : offset;
});
const weekdayMap = lazy(() => {
    const map = {};
    ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"].forEach((key, idx) => (map[key] = idx));
    return map;
});
const calcTimeZoneOffset = (dtf, dt) => {
    const parts = dtf.formatToParts(dt);
    const index = {};
    parts.forEach(v => (index[v.type] = v.value));
    // difference of days:
    let day = (7 + dt.getUTCDay() - weekdayMap()[index.weekday]) % 7;
    if (day > 3)
        day -= 7;
    // difference of hours: some locales use h24
    const hour = dt.getUTCHours() - (index.hour % 24);
    // difference of minutes:
    const minutes = dt.getUTCMinutes() - index.minute;
    // difference in minutes:
    return -((day * 24 + hour) * 60 + minutes);
};
const getTZF = cached(tz => {
    if (!/\//.test(tz)) {
        const fixed = parseTZ(tz);
        return _ => fixed;
    }
    // cache latest results
    let cache = {};
    let count = 0;
    let dtf;
    return ms => {
        // time zone offset never changes within every 15 minutes
        const minute15 = Math.floor(ms / 900000 /* d.MINUTE15 */);
        // check cached result
        let offset = cache && cache[minute15];
        if (offset != null)
            return offset;
        // reset all cache simply at every 24 x 4 times
        if (count++ > 96) {
            cache = {};
            count = 0;
        }
        const dt = new Date(ms);
        // DateTimeFormat is much slow
        if (!dtf) {
            dtf = new Intl.DateTimeFormat("en-US", { timeZone: tz, hour12: false, weekday: "short", hour: "numeric", minute: "numeric" });
        }
        offset = calcTimeZoneOffset(dtf, dt);
        // fallback to local time zone
        if (offset == null) {
            offset = -dt.getTimezoneOffset();
        }
        // write to cache
        cache[minute15] = offset;
        return offset;
    };
});

class DateTZ extends DateUTC {
    constructor(dt, timezone) {
        const t = +dt;
        const tzf = getTZF(timezone);
        const tzo = tzf(t);
        super(new Date(t + tzo * 60000 /* d.MINUTE */));
        this.t = t;
        this.tzf = tzf;
        this.tzo = tzo;
    }
    valueOf() {
        return this.t;
    }
    setTime(msec) {
        const tzo = this.tzo = this.tzf(msec);
        this.dt.setTime(+msec + tzo * 60000 /* d.MINUTE */);
        return this.t = msec;
    }
    getTimezoneOffset() {
        return -this.tzo;
    }
}
const tzPlugin = (Parent) => {
    return class CDateTZ extends Parent {
        tz(timezone) {
            const out = this.inherit();
            out.x.rw = (dt) => new DateTZ(new Date(+dt), timezone);
            return out;
        }
    };
};

/**
 * build an on-demand Handlers for the language specified
 */
const makeLocale = (lang) => {
    // build a handler function which converts from Date to string
    const makeHandler = (options, keyFn, partType) => {
        let format;
        const handler = (dt) => {
            // force UTC instead of local time
            const offset = dt.getTimezoneOffset();
            const utc = new Date(+dt - offset * 60000 /* d.MINUTE */);
            // cached DateTimeFormat instance
            if (!format)
                format = new Intl.DateTimeFormat(lang, options);
            // pickup only the single part of parts
            if (partType) {
                const part = format.formatToParts(utc).find(v => v.type === partType);
                return part && part.value;
            }
            // stringify
            const text = format.format(utc);
            // remove "UTC" string for some cases given
            if (text)
                return text.replace(/\s+UTC$/, "");
        };
        // on-demand use only
        if (!keyFn)
            return handler;
        // cached results
        const cache = {};
        return dt => {
            const key = keyFn(dt);
            return cache[key] || (cache[key] = handler(dt));
        };
    };
    // keyFn
    const getDay = (dt) => dt.getDay();
    const getMonth = (dt) => dt.getMonth();
    // Handlers
    return {
        "%a": makeHandler(styleOptions.a, getDay),
        "%A": makeHandler(styleOptions.A, getDay),
        "%b": makeHandler(styleOptions.b, getMonth),
        "%B": makeHandler(styleOptions.B, getMonth),
        "%c": makeHandler(styleOptions.c),
        "%p": makeHandler(styleOptions.p, dt => dt.getHours(), "dayPeriod"),
        "%r": makeHandler(styleOptions.r),
        "%x": makeHandler(styleOptions.x),
        "%X": makeHandler(styleOptions.X),
    };
};
const getLocaleOptions = () => {
    const digits = "2-digit";
    const medium = "medium";
    const numeric = "numeric";
    const short = "short";
    const long = "long";
    // Note: "timeZoneName" parameter is not allowed here!
    const options = {
        a: { weekday: short },
        A: { weekday: long },
        b: { month: short },
        B: { month: long },
        c: { weekday: short, year: numeric, month: short, day: numeric, hour: digits, minute: digits, second: digits },
        p: { hour: numeric, hour12: true },
        r: { timeStyle: medium, hour12: true },
        x: { dateStyle: short },
        X: { timeStyle: medium }, // hour12: default
    };
    // force UTC instead of local time
    Object.keys(options).forEach((key) => options[key].timeZone = "UTC");
    return options;
};
const styleOptions = getLocaleOptions();
const getLocale = cached(makeLocale);
const localePlugin = (Parent) => {
    return class CDateLocale extends Parent {
        locale(lang) {
            return this.handler(getLocale(lang));
        }
    };
};

class CDateCore {
    /**
     * the constructor
     */
    constructor(t, x) {
        this.t = t;
        if ("number" !== typeof t) {
            this.d = t;
        }
        this.x = x || {};
    }
    /**
     * creates another CDate object
     */
    cdate(dt) {
        if (dt == null) {
            dt = new Date(); // now
        }
        else if ("string" === typeof dt) {
            dt = new Date(dt); // parse ISO string
        }
        else {
            dt = new Date(+dt); // number or DateLike
        }
        return this.create(dt);
    }
    /**
     * returns a read-write version of DateLike for manipulation
     */
    rw() {
        const t = +this.t;
        const rw = this.x.rw;
        return rw ? rw(t) : new Date(t);
    }
    /**
     * returns a read-only version of DateLike for displaying
     */
    ro() {
        return this.d || (this.d = this.rw());
    }
    /**
     * returns milliseconds since the epoch
     */
    valueOf() {
        return +this.ro();
    }
    /**
     * returns a raw Date object
     */
    toDate() {
        return new Date(+this);
    }
    /**
     * returns a JSON string
     */
    toJSON() {
        return this.toDate().toJSON();
    }
    plugin(fn) {
        const CDateClass = this.constructor;
        const CDateX = fn(CDateClass) || CDateClass;
        return new CDateX(this.t, this.x);
    }
    /**
     * creates another CDate object with the DateLike given
     */
    create(dt) {
        return new this.constructor(dt, this.x);
    }
    /**
     * clones the CDate object
     */
    inherit() {
        const out = this.create(+this);
        // x is readonly
        out.x = Object.create(out.x);
        return out;
    }
}
const root = new CDateCore(0, {})
    .plugin(formatPlugin)
    .plugin(calcPlugin)
    .plugin(localePlugin)
    .plugin(utcPlugin)
    .plugin(tzPlugin);
const cdate = (dt) => root.cdate(dt);

exports.cdate = cdate;
exports.strftime = strftime;
